@@ -4,7 +4,7 @@ use 5.008_001;
 use strict;
 use warnings;
 
-our $VERSION = '0.1029';
+our $VERSION = '0.1030';
 
 use Text::Xslate::Util qw($DEBUG html_escape escaped_string);
 
@@ -43,6 +43,11 @@ BEGIN {
     *_DUMP_LOAD_FILE = sub(){ $dump_load_file };
 
     *_ST_MTIME = sub() { 9 }; # see perldoc -f stat
+
+    my $cache_dir = File::Spec->catfile(
+        $ENV{HOME} || File::Spec->tmpdir,
+        '.xslate_cache');
+    *_DEFAULT_CACHE_DIR = sub() { $cache_dir };
 }
 
 my $IDENT   = qr/(?: [a-zA-Z_][a-zA-Z0-9_\@]* )/xms;
@@ -52,51 +57,68 @@ my $XSLATE_MAGIC = qq{.xslate "%s - %s - %s - %s - %s"\n};
 
 sub compiler_class() { 'Text::Xslate::Compiler' }
 
+sub options { # overridable
+    my($class) = @_;
+    return {
+        # name => default
+        suffix      => '.tx',
+        path        => ['.'],
+        compiler    => $class->compiler_class,
+        input_layer => ':utf8',
+        syntax      => 'Kolon',
+        escape      => 'html',
+        cache       => 1,
+        cache_dir   => _DEFAULT_CACHE_DIR,
+        module      => undef,
+        function    => undef,
+
+        verbose      => 1,
+        warn_handler => undef,
+        die_handler  => undef,
+    };
+}
+
 sub new {
     my $class = shift;
     my %args  = (@_ == 1 ? %{$_[0]} : @_);
 
-    # options
+    my $options = $class->options;
+    my $used    = 0;
+    my $nargs   = scalar keys %args;
+    while(my $key = each %{$options}) {
+        if(exists $args{$key}) {
+            $used++;
+        }
+        if(!defined($args{$key}) && defined($options->{$key})) {
+            $args{$key} = $options->{$key};
+        }
+    }
 
-    defined($args{suffix})      or $args{suffix}      = '.tx';
-    defined($args{path})        or $args{path}        = [ '.' ];
-    defined($args{input_layer}) or $args{input_layer} = ':utf8';
-    defined($args{compiler})    or $args{compiler}    = $class->compiler_class;
-    defined($args{syntax})      or $args{syntax}      = 'Kolon';
-    defined($args{escape})      or $args{escape}      = 'html'; # or 'none'
-    defined($args{cache})       or $args{cache}       = 1; # 0, 1, 2
-    defined($args{cache_dir})   or $args{cache_dir}   = File::Spec->catfile(
-        $ENV{HOME} || File::Spec->tmpdir, '.xslate_cache',
-    );
+    if($used != $nargs) {
+        my @unknowns = grep { !exists $options->{$_} } keys %args;
+        warnings::warnif(misc => "$class: Unknown option(s): " . join ' ', @unknowns);
+    }
 
     my %funcs;
-
-    if(defined $args{import}) {
-        Carp::carp("'import' option has been renamed to 'module'"
-            . " because of the confliction with Perl's import() method."
-            . " Use 'module' instead");
-        %funcs = import_from(@{$args{import}});
-    }
     if(defined $args{module}) {
         %funcs = import_from(@{$args{module}});
     }
 
     # function => { ... } overrides imported functions
-    if(my $funcs_ref = $args{function}) {
+    if(defined(my $funcs_ref = $args{function})) {
         while(my($name, $body) = each %{$funcs_ref}) {
             $funcs{$name} = $body;
         }
     }
 
+    # the following functions are not overridable
     foreach my $builtin(qw(raw html dump)) {
         if(exists $funcs{$builtin}) {
             warnings::warnif(redefine =>
-                "You cannot redefine builtin function '$builtin',"
+                "$class: You cannot redefine builtin function '$builtin',"
                 . " because it is embeded in the engine");
         }
     }
-
-    # the following functions are not overridable
     $funcs{raw}  = \&Text::Xslate::Util::escaped_string;
     $funcs{html} = \&Text::Xslate::Util::html_escape;
     $funcs{dump} = \&Text::Xslate::Util::p;
@@ -110,14 +132,7 @@ sub new {
     # internal data
     $args{template} = {};
 
-    my $self = bless \%args, $class;
-
-    if(defined $args{string}) {
-        Carp::carp('"string" option has been deprecated. Use render_string($string, \%vars) instead');
-        $self->load_string($args{string});
-    }
-
-    return $self;
+    return bless \%args, $class;
 }
 
 sub load_string { # for <input>
@@ -126,7 +141,7 @@ sub load_string { # for <input>
         $self->_error("LoadError: Template string is not given");
     }
     $self->{string} = $string;
-    my $asm = $self->_compiler->compile($string);
+    my $asm = $self->compile($string);
     $self->_assemble($asm, undef, undef, undef, undef);
     return $asm;
 }
@@ -242,7 +257,7 @@ sub load_file {
         }
     }
     else {
-        $asm = $self->_compiler->compile($string,
+        $asm = $self->compile($string,
             file     => $file,
             fullpath => $fullpath,
         );
@@ -255,7 +270,8 @@ sub load_file {
                 File::Path::mkpath($cachedir);
             }
 
-            if(open my($out), '>:raw:utf8', $cachepath) {
+            # use input_layer for caches
+            if(open my($out), '>' . $self->{input_layer}, $cachepath) {
                 print $out $self->serialize($asm, $fullpath);
 
                 if(!close $out) {
@@ -318,6 +334,11 @@ sub _compiler {
     return $compiler;
 }
 
+sub compile {
+    my $self = shift;
+    return $self->_compiler->compile(@_);
+}
+
 sub deserialize {
     my($self, $assembly) = @_;
 
@@ -365,7 +386,7 @@ Text::Xslate - High performance template engine
 
 =head1 VERSION
 
-This document describes Text::Xslate version 0.1029.
+This document describes Text::Xslate version 0.1030.
 
 =head1 SYNOPSIS
 
@@ -524,34 +545,37 @@ You B<should> specify this option on productions.
 
 Specifies functions, which may be called as C<f($arg)> or C<$arg | f>.
 
-You can also define methods with pseudo type names: C<scalar>, C<array>,
-and C<hash>. For example:
-
-    my $tx = Text::Xslate->new(
-        function => {
-            'scalar::some_method' => sub { my($scalar)    = @_; ... },
-            'array::some_method'  => sub { my($array_ref) = @_; ... },
-            'hash::some_method'   => sub { my($hash_ref)  = @_; ... },
-        },
-    );
+Note that builtin methods are overridable, while builtin filters,
+namely C<raw>, C<html> and C<dump>, are not.
 
 =item C<< module => [$module => ?\@import_args, ...] >>
 
-Imports functions from I<$module>. I<@import_args> is optional.
+Imports functions from I<$module>, which may be a function-based or bridge module.
+I<@import_args> is optional.
 
 For example:
 
+    # for function-based modules
     my $tx = Text::Xslate->new(
-        module => ['Data::Dumper'], # use Data::Dumper
+        module => ['Time::Piece'],
     );
     print $tx->render_string(
-        '<: Dumper($x) :>',
-        { x => [42] },
-    );
-    # => $VAR = [42]
+        '<: localtime($x).strftime() :>',
+        { x => time() },
+    ); # => Wed, 09 Jun 2010 10:22:06 JST
 
-You can use function based modules with the C<module> option, and also can invoke
-object methods in templates. Thus, Xslate doesn't require the namespaces for plugins.
+    # for bridge modules
+    my $tx = Text::Xslate->new(
+        module => ['SomeModule::Bridge::Xslate'],
+    );
+    print $tx->render_string(
+        '<: $x.some_method() :>',
+        { x => time() },
+    );
+
+You can use function-based modules with the C<module> option, and also can
+invoke any object methods in templates. Thus, Xslate doesn't require the
+specific namespaces for plugins.
 
 =item C<< input_layer => $perliolayers // ':utf8' >>
 
@@ -675,12 +699,12 @@ which is explained in L<Text::Xslate::Syntax::TTerse>.
 
 =head1 NOTES
 
-There are common notes in the Xslate virtual machine.
+There are common notes in Xslate.
 
-=head2 Nil handling
+=head2 Nil/undef handling
 
-Note that nil handling is different from Perl's. Basically it does nothing,
-but C<< verbose => 2 >> will produce warnings for it.
+Note that nil (i.e. C<undef> in Perl) handling is different from Perl's.
+Basically it does nothing, but C<< verbose => 2 >> will produce warnings for it.
 
 =over
 
@@ -706,26 +730,6 @@ C<< $var == nil >> returns true if and only if I<$var> is nil.
 
 =back
 
-=head2 Automatic semicolon insertion
-
-The Xslate tokenizer automatically inserts semicolons at the end of the line
-codes. Currently this mechanism is not so smart, which could cause problems:
-
-For example, the following Kolon template causes syntax errors.
-
-    : my $foo = {
-    :    bar => 42,
-    : };
-
-It must be:
-
-    <: my $foo = {
-         bar => 42,
-       };
-    -:>
-
-This limitation should be resolved in a future.
-
 =head1 DEPENDENCIES
 
 Perl 5.8.1 or later.
